//-----------------------------------------------------------------------------
// MK3_2560_TWI
// Version 1.2
// 22.03.2013 / UB
// "www.Mikrocontroller-4U.de"
// mc-4u(@)t-online.de
//-----------------------------------------------------------------------------
// Fuer das myAVR MK3 Board (oder MK2 Board)
// mit ATMega-2560 @ 16MHz (oder ATMega8) 
// TWI-Funktionen
//-----------------------------------------------------------------------------
// das File muss im Ordner
// "myAVR/WinAVR/AVR/Include/UB_MK3_2560/"
// liegen und mit #include <UB_MK3_2560/MK3_2560_TWI.h>
// eingebunden werden
//-----------------------------------------------------------------------------
// Belegt : @MK3 (ATMega2560)
//          Port-D (Bit0 bis Bit1)
// Belegt : @MK2 (ATMega8)
//          Port-C (Bit4 bis Bit5)
//
// Hinweis :
//
// > die DEFINE-Schalter "TWI_HW_MODE" und "TWI_QUARZ_TYP"
//   muessen zur Hardware passen !! 
//-----------------------------------------------------------------------------


#ifndef MK3_2560_TWI    // Wichtig !!
#define MK3_2560_TWI 1  // damit wird das File nur einmal eingebunden



//-----------------------------------------------------------------------------
// Hardware auswahl
//-----------------------------------------------------------------------------
#define TWI_HW_MODE    0       // 0=MK3-Board, 1=MK1/MK2-Board/mySmartControl
#define TWI_QUARZ_TYP  0       // 0=16MHz, 1=8MHz, 2=3.6864MHz



//-----------------------------------------------------------------------------
// INFOS zur Benutzung :
// 01. Funktion "MK3_INIT_TWI" starten
// 02. Funktion "MK3_TWI_START" zum Starten der TWI-Kommunkation
// 03. Funktion "MK3_TWI_STOP" zum beenden der TWI-Kommunkation
// 04. Funktion "MK3_TWI_SEND" zum senden eines Bytes an den Slave
// 05. Funktion "MK3_TWI_READ" zum lesen eines Bytes vom Slave
// 06. Funktion "MK3_TWI_WAIT" fuer eine kleine Pause
//-----------------------------------------------------------------------------



//-----------------------------------------------------------------------------
// Funktionen die vom MAIN benutzt werden koennen
//-----------------------------------------------------------------------------
void MK3_INIT_TWI(void);		     // zum initialis. vom TWI-Bus
bool MK3_TWI_START(void);		     // zum starten der Kommunikation
void MK3_TWI_STOP(void);		     // zum beenden der Kommunikation
char MK3_TWI_SEND(char wert, bool send_ack); // sendet ein byte zum Slave
char MK3_TWI_READ(bool send_ack);            // liest ein byte vom Slave
void MK3_TWI_WAIT(char p);                   // kleine Pause
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Enumerationen fuer die Funktioen
//-----------------------------------------------------------------------------
// keine


//-----------------------------------------------------------------------------
// Port-Pin definition
//-----------------------------------------------------------------------------
#if TWI_HW_MODE==0
  // Hardware = myAVR MK3-Board
  #define MK3_DDR_TWI DDRD	// DDR-Register des TWI = PORT-D

  #define MK3_TWI_SCL 0x01    	// CLOCK von TWI = Bit0
  #define MK3_TWI_SDA 0x02      // DATA von TWI  = Bit1
#else
  // Hardware = myAVR MK1/MK2-Board + mySmartControl 
  #define MK3_DDR_TWI DDRC	// DDR-Register des TWI = PORT-C

  #define MK3_TWI_SCL 0x20    	// CLOCK von TWI = Bit5
  #define MK3_TWI_SDA 0x10      // DATA von TWI  = Bit4
#endif

//-----------------------------------------------------------------------------
// Register definition TWCR
//-----------------------------------------------------------------------------
#define BIT_TWEN  0x04          // TWEN  = BIT-2
#define BIT_TWSTO 0x10          // TESTO = BIT-4
#define BIT_TWSTA 0x20          // TWSTA = BIT-5
#define BIT_TWEA  0x40          // TWEA  = BIT-6
#define BIT_TWINT 0x80          // TWINT = BIT-7

//-----------------------------------------------------------------------------
// Register definition TWSR
//-----------------------------------------------------------------------------
#define MASKE_TWSR     0xF8     // Bit7 bis Bit3
#define TWSR_START_OK  0x08     // Erfolgreiche Start-Kennung
#define TWSR_SLAVE_OK  0x18     // ACK-Kennung vom Slave

//-----------------------------------------------------------------------------
// Defines
//-----------------------------------------------------------------------------
#define TWI_ERR_MAX   1000      // Fehlercounter (nur damit keine Endlosloop)

//-----------------------------------------------------------------------------
// Globale interne Variablen
//-----------------------------------------------------------------------------
bool twi_init=false;

//-----------------------------------------------------------------------------
void MK3_INIT_TWI(void) {	// zum initialisieren der TWI-Schnittstelle

  if(twi_init==false) {
    // die initialisierung wird nur einmal gemacht
    twi_init=true;

    // init der Port-Pins eigentlich nicht notwendig  
    MK3_DDR_TWI&=~MK3_TWI_SCL;	// SCL als Eingang
    MK3_DDR_TWI&=~MK3_TWI_SDA;	// SDA als Eingang


    // TWI_FRQ wird hier auf 100kHz eingestellt
    #if TWI_QUARZ_TYP==0
      // 16MHz
      TWBR=72;			// TWI_FRQ = 100kHz TWBR=((Quarz/TWI_FRQ)-16)/2
    #elif TWI_QUARZ_TYP==1
      // 8MHz
      TWBR=32;			// TWI_FRQ = 100kHz TWBR=((Quarz/TWI_FRQ)-16)/2
    #elif TWI_QUARZ_TYP==2
      // 3.6864MHz
	  TWBR=10;			// TWI_FRQ = 100kHz TWBR=((Quarz/TWI_FRQ)-16)/2
    #endif

    TWCR=0x00;
    TWSR=0x00;			// Vorteiler = 1
    TWDR=0xFF;                  // dummy Inhalt
  
    // TWI einschalten
    TWCR|=BIT_TWEN;		// TWI enable
    TWCR|=BIT_TWINT;		// Interrupt Flag loeschen
  } 
}


//-----------------------------------------------------------------------------
bool MK3_TWI_START(void) { // zum starten der Kommunikation
  // bei einem Fehler wird FALSE zurueckgegeben
  
  bool ret_wert=true;
  unsigned int p;
  
  TWCR=BIT_TWEN;		// TWI enable
  TWCR|=BIT_TWSTA;              // TWSTA
  TWCR|=BIT_TWINT;		// Interrupt Flag loeschen
  
  p=0;
  while( !(TWCR & BIT_TWINT)) {
    // warten bis gesendet
    p++;
    if(p>TWI_ERR_MAX) {
      ret_wert=false; 
      break; // abbruch bei fehler
    }
  }
  
  if((ret_wert==true) && ((TWSR&MASKE_TWSR)!=TWSR_START_OK)) {
    // Start-Kennung wurde nicht erfolgreich gesendet
    ret_wert=false; 
  }
  
  return(ret_wert);
}

//-----------------------------------------------------------------------------
void MK3_TWI_STOP(void) { // zum beenden der Kommunikation

  TWCR=BIT_TWEN;		// TWI enable
  TWCR|=BIT_TWSTO;              // TWSTO
  TWCR|=BIT_TWINT;		// Interrupt Flag loeschen
}

//-----------------------------------------------------------------------------
char MK3_TWI_SEND(char wert, bool send_ack) { // sendet ein byte zum slave
  // wenn alles OK -> wird der Inhalt von TWSR zurueckgegeben
  
  unsigned int p;
  char ret_wert=0;
  bool send_error=false;
  
  TWDR=wert;			// Daten zum senden

  TWCR=BIT_TWEN;		// TWI enable
  
  if(send_ack==true) {
    TWCR|=BIT_TWEA;		// Acknoledge setzen
  }

  TWCR|=BIT_TWINT;		// Interrupt Flag loeschen

  p=0;
  while( !(TWCR & BIT_TWINT)) {
    // warten bis gesendet
    p++;
    if(p>TWI_ERR_MAX) {
      send_error=true;
      break; // abbruch bei fehler
    }    
  }
  
  if(send_error==false) {
    // Inhalt von TWSR zurueckgeben
    ret_wert=(TWSR&MASKE_TWSR);
  }
  
  return(ret_wert);
}

//-----------------------------------------------------------------------------
char MK3_TWI_READ(bool send_ack) { // liest ein byte vom slave
  // wenn alles OK -> wird der Inhalt von TWDR zurueckgegeben
  
  char ret_wert=0;
  unsigned int p;
  bool read_error=false;
  
  TWCR=BIT_TWEN;		// TWI enable

  if(send_ack==true) {
    TWCR|=BIT_TWEA;		// Acknoledge setzen
  }

  TWCR|=BIT_TWINT;		// Interrupt Flag loeschen
  
  p=0;
  while( !(TWCR & BIT_TWINT)) {
    // warten bis gesendet
    p++;
    if(p>TWI_ERR_MAX) {
      read_error=true;
      break; // abbruch bei fehler
    }    
  }
  
  if(read_error==false) {
    // Daten vom Slave zurueckgeben
    ret_wert=TWDR;
  }

  return(ret_wert);
}

//-----------------------------------------------------------------------------
void MK3_TWI_WAIT(char p) { // kleine pause
  unsigned int n=0;
  char m=0;

  for(m=0;m<=p;m++) {
    for(n=0;n<=1000;n++) { // 1000 @ 16MHz = ca. 0,5ms
      // kleine Pause
      __asm volatile ("nop \n");  // mini Pause
    }
  }
}



#endif  // Ende von #define MK3_2560_TWI

